﻿/*
 * AppManager.cs
 * Copyright (c) 2009 kbinani
 *
 * This file is part of Boare.Cadencii.
 *
 * Boare.Cadencii is free software; you can redistribute it and/or
 * modify it under the terms of the GPLv3 License.
 *
 * Boare.Cadencii is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
using System;
using System.CodeDom.Compiler;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Reflection;
using System.Runtime.Serialization.Formatters.Binary;
using System.Windows.Forms;
using System.Xml.Serialization;
using Microsoft.CSharp;

using Boare.Lib.AppUtil;
using Boare.Lib.Vsq;
using bocoree;
/*
-コントロールカーブのテンプレートの登録＆貼付け機能
-選択したノートを指定したクロック数分一括シフトする機能
-前回レンダリングしたWAVEファイルをリサイクルする（起動ごとにレンダリングするのは面倒？）
-横軸を変更するショートカットキー
-音符の時間位置を秒指定で入力できるモード
-コントロールカーブのリアルタイムかつアナログ的な入力（ジョイスティックorゲームコントローラ）
-コントロールカーブエディタを幾つか積めるような機能
 */
namespace Boare.Cadencii {

    using boolean = System.Boolean;
    using Integer = System.Int32;

    public delegate void BSimpleDelegate<T>( T arg );

    public static class AppManager {
        /// <summary>
        /// 鍵盤の表示幅(pixel)
        /// </summary>
        public const int KEY_LENGTH = 68;
        private const String CONFIG_FILE_NAME = "config.xml";
        private const String CONFIG_DIR_NAME = "Cadencii";
        /// <summary>
        /// OSのクリップボードに貼り付ける文字列の接頭辞．これがついていた場合，クリップボードの文字列をCadenciiが使用できると判断する．
        /// </summary>
        private const String CLIP_PREFIX = "CADENCIIOBJ";

        /// <summary>
        /// エディタの設定
        /// </summary>
        public static EditorConfig editorConfig = new EditorConfig();
        /// <summary>
        /// AttachedCurve用のシリアライザ
        /// </summary>
        public static XmlSerializer xmlSerializerListBezierCurves = new XmlSerializer( typeof( AttachedCurve ) );
        public static Font baseFont8 = null;
        public static Font baseFont9 = null;
        /// <summary>
        /// ピアノロールの歌詞の描画に使用されるフォント。
        /// </summary>
        public static Font baseFont10 = null;
        /// <summary>
        /// ピアノロールの歌詞の描画に使用されるフォント。（発音記号固定の物の場合）
        /// </summary>
        public static Font baseFont10Bold = null;
        /// <summary>
        /// 歌詞を音符の（高さ方向の）真ん中に描画するためのオフセット
        /// </summary>
        public static int baseFont10OffsetHeight = 0;
        public static int baseFont8OffsetHeight = 0;
        public static int baseFont9OffsetHeight = 0;
        public static PropertyPanel propertyPanel;
        public static FormNoteProperty propertyWindow;

        #region Static Readonly Fields
        public static readonly Color[] HILIGHT = new Color[] { 
            Color.FromArgb( 181, 220, 16 ),
            Color.FromArgb( 231, 244, 49 ),
            Color.FromArgb( 252, 230, 29 ),
            Color.FromArgb( 247, 171, 20 ),
            Color.FromArgb( 249, 94, 17 ),
            Color.FromArgb( 234, 27, 47 ),
            Color.FromArgb( 175, 20, 80 ),
            Color.FromArgb( 183, 24, 149 ),
            Color.FromArgb( 105, 22, 158 ),
            Color.FromArgb( 22, 36, 163 ),
            Color.FromArgb( 37, 121, 204 ),
            Color.FromArgb( 29, 179, 219 ),
            Color.FromArgb( 24, 239, 239 ),
            Color.FromArgb( 25, 206, 175 ),
            Color.FromArgb( 23, 160, 134 ),
            Color.FromArgb( 79, 181, 21 ) };
        public static readonly Color[] RENDER = new Color[]{
            Color.FromArgb( 19, 143, 52 ),
            Color.FromArgb( 158, 154, 18 ),
            Color.FromArgb( 160, 143, 23 ),
            Color.FromArgb( 145, 98, 15 ),
            Color.FromArgb( 142, 52, 12 ),
            Color.FromArgb( 142, 19, 37 ),
            Color.FromArgb( 96, 13, 47 ),
            Color.FromArgb( 117, 17, 98 ),
            Color.FromArgb( 62, 15, 99 ),
            Color.FromArgb( 13, 23, 84 ),
            Color.FromArgb( 25, 84, 132 ),
            Color.FromArgb( 20, 119, 142 ),
            Color.FromArgb( 19, 142, 139 ),
            Color.FromArgb( 17, 122, 102 ),
            Color.FromArgb( 13, 86, 72 ),
            Color.FromArgb( 43, 91, 12 ) };
        public static readonly String[] USINGS = new String[] { "using System;",
                                             "using System.IO;",
                                             "using Boare.Lib.Vsq;",
                                             "using Boare.Cadencii;",
                                             "using bocoree;",
                                             "using Boare.Lib.Media;",
                                             "using Boare.Lib.AppUtil;",
                                             "using System.Windows.Forms;",
                                             "using System.Collections.Generic;",
                                             "using System.Drawing;",
                                             "using System.Text;",
                                             "using System.Xml.Serialization;" };
        public static readonly Vector<Keys> SHORTCUT_ACCEPTABLE = new Vector<Keys>( new Keys[]{
            Keys.A,
            Keys.B,
            Keys.C,
            Keys.D,
            Keys.D0,
            Keys.D1,
            Keys.D2,
            Keys.D3,
            Keys.D4,
            Keys.D5,
            Keys.D6,
            Keys.D7,
            Keys.D8,
            Keys.D9,
            Keys.Down,
            Keys.E,
            Keys.F,
            Keys.F1,
            Keys.F2,
            Keys.F3,
            Keys.F4,
            Keys.F5,
            Keys.F6,
            Keys.F7,
            Keys.F8,
            Keys.F9,
            Keys.F10,
            Keys.F11,
            Keys.F12,
            Keys.F13,
            Keys.F14,
            Keys.F15,
            Keys.F16,
            Keys.F17,
            Keys.F18,
            Keys.F19,
            Keys.F20,
            Keys.F21,
            Keys.F22,
            Keys.F23,
            Keys.F24,
            Keys.G,
            Keys.H,
            Keys.I,
            Keys.J,
            Keys.K,
            Keys.L,
            Keys.Left,
            Keys.M,
            Keys.N,
            Keys.NumPad0,
            Keys.NumPad1,
            Keys.NumPad2,
            Keys.NumPad3,
            Keys.NumPad4,
            Keys.NumPad5,
            Keys.NumPad6,
            Keys.NumPad7,
            Keys.NumPad8,
            Keys.NumPad9,
            Keys.O,
            Keys.P,
            Keys.PageDown,
            Keys.PageUp,
            Keys.Q,
            Keys.R,
            Keys.Right,
            Keys.S,
            Keys.Space,
            Keys.T,
            Keys.U,
            Keys.Up,
            Keys.V,
            Keys.W,
            Keys.X,
            Keys.Y,
            Keys.Z } );
        public static readonly CurveType[] CURVE_USAGE = new CurveType[]{ CurveType.DYN,
                                                                          CurveType.BRE,
                                                                          CurveType.BRI,
                                                                          CurveType.CLE,
                                                                          CurveType.OPE,
                                                                          CurveType.GEN,
                                                                          CurveType.POR,
                                                                          CurveType.PIT,
                                                                          CurveType.PBS,
                                                                          CurveType.VibratoRate,
                                                                          CurveType.VibratoDepth,
                                                                          CurveType.harmonics,
                                                                          CurveType.fx2depth,
                                                                          CurveType.reso1amp,
                                                                          CurveType.reso1bw,
                                                                          CurveType.reso1freq,
                                                                          CurveType.reso2amp,
                                                                          CurveType.reso2bw,
                                                                          CurveType.reso2freq,
                                                                          CurveType.reso3amp,
                                                                          CurveType.reso3bw,
                                                                          CurveType.reso3freq,
                                                                          CurveType.reso4amp,
                                                                          CurveType.reso4bw,
                                                                          CurveType.reso4freq };
        #endregion

        #region Private Static Fields
        private static int s_base_tempo = 480000;
        private static SolidBrush s_hilight_brush = new SolidBrush( Color.CornflowerBlue );
        private static object s_locker;
        //private static XmlSerializer s_serizlizer = null;
        #endregion

        #region Imported from EditorManager
        /// <summary>
        /// 現在表示されているピアノロール画面の右上の、仮想スクリーン上座標で見たときのy座標(pixel)
        /// </summary>
        public static int startToDrawY = 0;
#if !TREECOM
        private static VsqFileEx s_vsq;
#endif
        private static String s_file = "";
        private static int s_selected = 1;
        private static int s_current_clock = 0;
        private static boolean s_playing = false;
        private static boolean s_repeat_mode = false;
        private static boolean s_grid_visible = false;
        private static EditMode s_edit_mode = EditMode.NONE;
        /// <summary>
        /// トラックのレンダリングが必要かどうかを表すフラグ
        /// </summary>
        private static boolean[] s_render_required = new boolean[16] { false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false };
        /// <summary>
        /// トラックのオーバーレイ表示
        /// </summary>
        private static boolean s_overlay = true;
        private static SelectedEventList s_selected_event = new SelectedEventList();
        /// <summary>
        /// 選択されているテンポ変更イベントのリスト
        /// </summary>
        private static TreeMap<Integer, SelectedTempoEntry> s_selected_tempo = new TreeMap<Integer, SelectedTempoEntry>();
        private static int s_last_selected_tempo = -1;
        /// <summary>
        /// 選択されている拍子変更イベントのリスト
        /// </summary>
        private static TreeMap<Integer, SelectedTimesigEntry> s_selected_timesig = new TreeMap<Integer, SelectedTimesigEntry>();
        private static int s_last_selected_timesig = -1;
        private static EditTool s_selected_tool = EditTool.PENCIL;
#if !TREECOM
        private static Vector<ICommand> s_commands = new Vector<ICommand>();
        private static int s_command_position = -1;
#endif
        /// <summary>
        /// 選択されているベジエ点のリスト
        /// </summary>
        private static Vector<SelectedBezierPoint> s_selected_bezier = new Vector<SelectedBezierPoint>();
        /// <summary>
        /// 最後に選択されたベジエ点
        /// </summary>
        private static SelectedBezierPoint s_last_selected_bezier = new SelectedBezierPoint();
        /// <summary>
        /// 現在の演奏カーソルの位置(m_current_clockと意味は同じ。CurrentClockが変更されると、自動で更新される)
        /// </summary>
        private static PlayPositionSpecifier s_current_play_position;

        /// <summary>
        /// SelectedRegionが有効かどうかを表すフラグ
        /// </summary>
        public static boolean selectedRegionEnabled = false;
        /// <summary>
        /// Ctrlキーを押しながらのマウスドラッグ操作による選択が行われた範囲
        /// </summary>
        public static SelectedRegion selectedRegion;
        /// <summary>
        /// 自動ノーマライズモードかどうかを表す値を取得、または設定します。
        /// </summary> 
        public static boolean autoNormalize = false;
        /// <summary>
        /// エンドマーカーの位置(clock)
        /// </summary>
        public static int endMarker = 0;
        public static boolean endMarkerEnabled = false;
        /// <summary>
        /// x方向の表示倍率(pixel/clock)
        /// </summary>
        public static float scaleX = 0.1f;
        /// <summary>
        /// スタートマーカーの位置(clock)
        /// </summary>
        public static int startMarker = 0;
        /// <summary>
        /// Bezierカーブ編集モードが有効かどうかを表す
        /// </summary>
        private static boolean s_is_curve_mode = false;
        public static boolean startMarkerEnabled = false;
        /// <summary>
        /// 再生時に自動スクロールするかどうか
        /// </summary>
        public static boolean autoScroll = true;
        /// <summary>
        /// 最初のバッファが書き込まれたかどうか
        /// </summary>
        public static boolean firstBufferWritten = false;
        /// <summary>
        /// リアルタイム再生で使用しようとしたVSTiが有効だったかどうか。
        /// </summary>
        public static boolean rendererAvailable = false;
        /// <summary>
        /// プレビュー再生が開始された時刻
        /// </summary>
        public static DateTime previewStartedTime;
        /// <summary>
        /// 画面左端位置での、仮想画面上の仮想画面左端から測ったピクセル数．FormMain.hScroll.ValueとFormMain.trackBar.Valueで決まる．
        /// </summary>
        public static int startToDrawX;
        public static String selectedPaletteTool = "";
        public static ScreenStatus lastScreenStatus = new ScreenStatus();
        private static String s_id = "";
        public static FormMain mainForm = null;

        /// <summary>
        /// グリッド線を表示するか否かを表す値が変更された時発生します
        /// </summary>
        public static event EventHandler GridVisibleChanged;
        public static event EventHandler PreviewStarted;
        public static event EventHandler PreviewAborted;
        public static event BSimpleDelegate<boolean> SelectedEventChanged;
        public static event EventHandler SelectedToolChanged;
        /// <summary>
        /// CurrentClockプロパティの値が変更された時発生します
        /// </summary>
        public static event EventHandler CurrentClockChanged;

        private const String TEMPDIR_NAME = "cadencii";

        /// <summary>
        /// pがrcの中にあるかどうかを判定します
        /// </summary>
        /// <param name="p"></param>
        /// <param name="rc"></param>
        /// <returns></returns>
        public static boolean isInRect( Point p, Rectangle rc ) {
            if ( rc.X <= p.X ) {
                if ( p.X <= rc.X + rc.Width ) {
                    if ( rc.Y <= p.Y ) {
                        if ( p.Y <= rc.Y + rc.Height ) {
                            return true;
                        }
                    }
                }
            }
            return false;
        }

        /// <summary>
        /// fontを使って文字を描画したとき，文字の縦方向の中心線と，描画時に指定した座標との間にどれだけのオフセットが生じるかを調べる
        /// </summary>
        /// <param name="font"></param>
        /// <returns></returns>
        public static unsafe int getStringOffset( Font font ) {
            SizeF str_size = measureString( "Qjp", font );
            if ( str_size.Width <= 0 || str_size.Height <= 0 ) {
                return 0;
            }
            int draw_pos = (int)(str_size.Height / 2);
            Bitmap test = new Bitmap( (int)str_size.Width * 2, (int)str_size.Height * 2, System.Drawing.Imaging.PixelFormat.Format24bppRgb );
            using ( Graphics g = Graphics.FromImage( test ) ) {
                g.DrawString(
                    "Qjp",
                    font,
                    Brushes.Black,
                    new PointF( str_size.Width / 2, draw_pos ) );
            }
            BitmapData bdat = test.LockBits(
                new Rectangle( 0, 0, test.Width, test.Height ),
                ImageLockMode.ReadOnly,
                PixelFormat.Format24bppRgb );
            int stride = bdat.Stride;
            byte* dat = (byte*)bdat.Scan0;
            boolean found = false;
            int firsty = draw_pos;
            for ( int y = 0; y < bdat.Height; y++ ) {
                int pos = y * stride - 3;
                for ( int x = 0; x < bdat.Width; x++ ) {
                    pos += 3;
                    if ( dat[pos] == 0 ) {
                        found = true;
                        break;
                    }
                }
                if ( found ) {
                    firsty = y;
                    break;
                }
            }
            found = false;
            int lasty = draw_pos;
            for ( int y = bdat.Height - 1; y >= 0; y-- ) {
                int pos = y * stride - 3;
                for ( int x = 0; x < bdat.Width; x++ ) {
                    pos++;
                    if ( dat[pos] == 0 ) {
                        found = true;
                        break;
                    }
                }
                if ( found ) {
                    lasty = y;
                    break;
                }
            }
            return (lasty + firsty) / 2 - draw_pos;
        }

        public static SizeF measureString( String item, Font font ) {
            using ( Bitmap test = new Bitmap( 10, 10 ) )
            using ( Graphics g = Graphics.FromImage( test ) ) {
                return g.MeasureString( item, font );
            }
        }

        /// <summary>
        /// 文字列itemをfontを用いて描画したとき、幅widthピクセルに収まるようにitemを調節したものを返します。
        /// 例えば"1 Voice"→"1 Voi..."ナド。
        /// </summary>
        /// <param name="item"></param>
        /// <param name="font"></param>
        /// <param name="width"></param>
        /// <returns></returns>
        public static String trimString( String item, Font font, int width ) {
            String edited = item;
            int delete_count = item.Length;
            while ( true ) {
                SizeF measured = measureString( edited, font );
                if ( measured.Width <= width ) {
                    return edited;
                }
                delete_count -= 1;
                if ( delete_count > 0 ) {
                    edited = item.Substring( 0, delete_count ) + "...";
                } else {
                    return edited;
                }
            }
        }


        public static void debugWriteLine( String message ) {
#if DEBUG
            Console.WriteLine( message );
#endif
        }

        /// <summary>
        /// FormMainを識別するID
        /// </summary>
        public static String getID() {
                return s_id;
        }

        public static String _( String id ) {
            return Boare.Lib.AppUtil.Messaging.GetMessage( id );
        }

        public static ScriptInvoker loadScript( String file ) {
#if DEBUG
            AppManager.debugWriteLine( "Common.LoadScript(String)" );
            AppManager.debugWriteLine( "    File.GetLastWriteTimeUtc( file )=" + File.GetLastWriteTimeUtc( file ) );
#endif
            ScriptInvoker ret = new ScriptInvoker();
            ret.ScriptFile = file;
            ret.FileTimestamp = File.GetLastWriteTimeUtc( file );
            // スクリプトの記述のうち、以下のリストに当てはまる部分は空文字に置換される
            String config_file = configFileNameFromScriptFileName( file );
            String script = "";
            using ( StreamReader sr = new StreamReader( file ) ) {
                script = sr.ReadToEnd();
            }
            foreach ( String s in AppManager.USINGS ) {
                script = script.Replace( s, "" );
            }

            String code = "";
            foreach ( String s in AppManager.USINGS ) {
                code += s;
            }
            code += "namespace Boare.CadenciiScript{";
            code += script;
            code += "}";
            ret.ErrorMessage = "";
            CompilerResults results;
            Assembly testAssembly = AppManager.compileScript( code, out results );
            if ( testAssembly == null | results == null ) {
                ret.ErrorMessage = "failed compiling";
                return ret;
            }
            if ( results.Errors.Count != 0 ) {
                for ( int i = 0; i < results.Errors.Count; i++ ) {
                    ret.ErrorMessage += _( "line" ) + ":" + results.Errors[i].Line + " " + results.Errors[i].ErrorText + Environment.NewLine;
                }
                if ( results.Errors.HasErrors ) {
                    return ret;
                }
            }

            if ( testAssembly != null ) {
                foreach ( Type implemented in testAssembly.GetTypes() ) {
                    MethodInfo tmi = implemented.GetMethod( "Edit", new Type[] { typeof( VsqFile ) } );
                    if ( tmi != null && tmi.IsStatic && tmi.IsPublic && tmi.ReturnType.Equals( typeof( boolean ) ) ) {
                        ret.ScriptType = implemented;
                        ret.EditVsqDelegate = (EditVsqScriptDelegate)Delegate.CreateDelegate( typeof( EditVsqScriptDelegate ), tmi );
                        ret.Serializer = new XmlStaticMemberSerializer( implemented );

                        if ( !File.Exists( config_file ) ) {
                            continue;
                        }

                        // 設定ファイルからDeserialize
                        using ( FileStream fs = new FileStream( config_file, FileMode.Open ) ) {
                            ret.Serializer.Deserialize( fs );
                        }
                    }
                    tmi = implemented.GetMethod( "Edit", new Type[] { typeof( VsqFileEx ) } );
                    if ( tmi != null && tmi.IsStatic && tmi.IsPublic && tmi.ReturnType.Equals( typeof( boolean ) ) ) {
                        ret.ScriptType = implemented;
                        ret.EditVsqDelegate2 = (EditVsqScriptDelegate2)Delegate.CreateDelegate( typeof( EditVsqScriptDelegate2 ), tmi );
                        ret.Serializer = new XmlStaticMemberSerializer( implemented );

                        if ( !File.Exists( config_file ) ) {
                            continue;
                        }

                        // 設定ファイルからDeserialize
                        using ( FileStream fs = new FileStream( config_file, FileMode.Open ) ) {
                            ret.Serializer.Deserialize( fs );
                        }
                    }
                }
            }
#if DEBUG
            Console.WriteLine( "(ret.EditVsqDelegate==null)=" + (ret.EditVsqDelegate == null) );
            Console.WriteLine( "(ret.EditVsqDelegate2==null)=" + (ret.EditVsqDelegate2 == null) );
#endif
            return ret;
        }

        /// <summary>
        /// スクリプトを実行します
        /// </summary>
        /// <param name="evsd"></param>
        public static boolean invokeScript( ScriptInvoker script_invoker ) {
            if ( script_invoker != null && (script_invoker.EditVsqDelegate != null || script_invoker.EditVsqDelegate2 != null) ) {
                try {
                    VsqFileEx work = (VsqFileEx)s_vsq.Clone();
                    boolean ret = false;
                    if ( script_invoker.EditVsqDelegate != null ) {
                        ret = script_invoker.EditVsqDelegate.Invoke( work );
                    } else if ( script_invoker.EditVsqDelegate2 != null ) {
                        ret = script_invoker.EditVsqDelegate2.Invoke( work );
                    }
                    if ( !ret ) {
                        MessageBox.Show( _( "Script aborted" ), "Cadencii", MessageBoxButtons.OK, MessageBoxIcon.Information );
                    } else {
                        CadenciiCommand run = VsqFileEx.generateCommandReplace( work );
                        register( s_vsq.executeCommand( run ) );
                    }
                    String config_file = configFileNameFromScriptFileName( script_invoker.ScriptFile );
                    using ( FileStream fs = new FileStream( config_file, FileMode.Create ) ) {
                        script_invoker.Serializer.Serialize( fs );
                    }
                    return ret;
                } catch ( Exception ex ) {
                    MessageBox.Show( _( "Script runtime error:" ) + " " + ex, _( "Error" ), MessageBoxButtons.OK, MessageBoxIcon.Information );
#if DEBUG
                    AppManager.debugWriteLine( "    " + ex );
#endif
                }
            } else {
                MessageBox.Show( _( "Script compilation failed." ), _( "Error" ), MessageBoxButtons.OK, MessageBoxIcon.Exclamation );
            }
            return false;
        }

        private static String configFileNameFromScriptFileName( String script_file ) {
            String dir = Path.Combine( AppManager.getApplicationDataPath(), "script" );
            if ( !Directory.Exists( dir ) ) {
                Directory.CreateDirectory( dir );
            }
            return Path.Combine( dir, Path.GetFileNameWithoutExtension( script_file ) + ".config" );
        }

        public static String getTempWaveDir() {
            String temp = Path.Combine( Path.GetTempPath(), TEMPDIR_NAME );
            String dir = Path.Combine( temp, getID() );
            if ( !Directory.Exists( temp ) ) {
                Directory.CreateDirectory( temp );
            }
            if ( !Directory.Exists( dir ) ) {
                Directory.CreateDirectory( dir );
            }
            return dir;
        }

        public static boolean isCurveMode() {
            return s_is_curve_mode;
        }

        public static void setCurveMode( boolean value ) {
            boolean old = s_is_curve_mode;
            s_is_curve_mode = value;
            if ( old != s_is_curve_mode && SelectedToolChanged != null ) {
                SelectedToolChanged( typeof( AppManager ), new EventArgs() );
            }
        }

#if !TREECOM
        /// <summary>
        /// コマンドの履歴を削除します
        /// </summary>
        public static void clearCommandBuffer() {
            s_commands.clear();
            s_command_position = -1;
        }

        /// <summary>
        /// アンドゥ処理を行います
        /// </summary>
        public static void undo() {
#if DEBUG
            AppManager.debugWriteLine( "CommonConfig.Undo()" );
#endif
            if ( isUndoAvailable() ) {
                ICommand run_src = s_commands.get( s_command_position );
                CadenciiCommand run = (CadenciiCommand)run_src;
#if DEBUG
                AppManager.debugWriteLine( "    command type=" + run.type );
#endif
                if ( run.vsqCommand != null ) {
                    if ( run.vsqCommand.Type == VsqCommandType.TRACK_DELETE ) {
                        int track = (int)run.vsqCommand.Args[0];
                        if ( track == getSelected() && track >= 2 ) {
                            setSelected( track - 1 );
                        }
                    }
                }
                ICommand inv = s_vsq.executeCommand( run );
                s_commands.set( s_command_position, inv );
                s_command_position--;
            }
        }

        /// <summary>
        /// リドゥ処理を行います
        /// </summary>
        public static void redo() {
#if DEBUG
            AppManager.debugWriteLine( "CommonConfig.Redo()" );
#endif
            if ( isRedoAvailable() ) {
                ICommand run_src = s_commands.get( s_command_position + 1 );
                CadenciiCommand run = (CadenciiCommand)run_src;
                if ( run.vsqCommand != null ) {
                    if ( run.vsqCommand.Type == VsqCommandType.TRACK_DELETE ) {
                        int track = (int)run.args[0];
                        if ( track == getSelected() && track >= 2 ) {
                            setSelected( track - 1 );
                        }
                    }
                }
                ICommand inv = s_vsq.executeCommand( run );
                s_commands.set( s_command_position + 1, inv );
                s_command_position++;
            }
        }

        /// <summary>
        /// コマンドバッファに指定されたコマンドを登録します
        /// </summary>
        /// <param name="command"></param>
        public static void register( ICommand command ) {
#if DEBUG
            AppManager.debugWriteLine( "EditorManager.Register; command=" + command );
            AppManager.debugWriteLine( "    m_commands.Count=" + s_commands.size() );
#endif
            if ( s_command_position == s_commands.size() - 1 ) {
                // 新しいコマンドバッファを追加する場合
                s_commands.add( command );
                s_command_position = s_commands.size() - 1;
            } else {
                // 既にあるコマンドバッファを上書きする場合
                s_commands.set( s_command_position + 1, command );
                for ( int i = s_commands.size() - 1; i >= s_command_position + 2; i-- ) {
                    s_commands.removeElementAt( i );
                }
                s_command_position++;
            }
        }

        /// <summary>
        /// リドゥ操作が可能かどうかを表すプロパティ
        /// </summary>
        public static boolean isRedoAvailable() {
            if ( s_commands.size() > 0 && 0 <= s_command_position + 1 && s_command_position + 1 < s_commands.size() ) {
                return true;
            } else {
                return false;
            }
        }

        /// <summary>
        /// アンドゥ操作が可能かどうかを表すプロパティ
        /// </summary>
        public static boolean isUndoAvailable() {
            if ( s_commands.size() > 0 && 0 <= s_command_position && s_command_position < s_commands.size() ) {
                return true;
            } else {
                return false;
            }
        }
#endif

        public static EditTool getSelectedTool() {
                return s_selected_tool;
        }

        public static void setSelectedTool( EditTool value ) {
            EditTool old = s_selected_tool;
            s_selected_tool = value;
            if ( old != s_selected_tool && SelectedToolChanged != null ) {
                SelectedToolChanged( typeof( AppManager ), new EventArgs() );
            }
        }

        #region SelectedBezier
        public static IEnumerable<SelectedBezierPoint> getSelectedBezierEnumerator() {
            for ( int i = 0; i < s_selected_bezier.size(); i++ ) {
                yield return s_selected_bezier.get( i );
            }
        }

        /// <summary>
        /// ベジエ点のリストに最後に追加された点の情報を取得します
        /// </summary>
        public static SelectedBezierPoint getLastSelectedBezier() {
            if ( s_last_selected_bezier.ChainID < 0 || s_last_selected_bezier.PointID < 0 ) {
                return null;
            } else {
                return s_last_selected_bezier;
            }
        }

        /// <summary>
        /// 選択されているベジエ点のリストに、アイテムを追加します
        /// </summary>
        /// <param name="selected">追加する点</param>
        public static void addSelectedBezier( SelectedBezierPoint selected ) {
            s_last_selected_bezier = selected;
            int index = -1;
            for ( int i = 0; i < s_selected_bezier.size(); i++ ) {
                if ( s_selected_bezier.get( i ).ChainID == selected.ChainID &&
                    s_selected_bezier.get( i ).PointID == selected.PointID ) {
                    index = i;
                    break;
                }
            }
            if ( index >= 0 ) {
                s_selected_bezier.set( index, selected );
            } else {
                s_selected_bezier.add( selected );
            }
        }

        /// <summary>
        /// 選択されているベジエ点のリストを初期化します
        /// </summary>
        public static void clearSelectedBezier() {
            s_selected_bezier.clear();
            s_last_selected_bezier.ChainID = -1;
            s_last_selected_bezier.PointID = -1;
        }
        #endregion

        #region SelectedTimesig
        /// <summary>
        /// 選択されている拍子変更イベントのリスト
        /// </summary>
        public static TreeMap<Integer, SelectedTimesigEntry> getSelectedTimesig() {
            return s_selected_timesig;
        }

        public static SelectedTimesigEntry getLastSelectedTimesig() {
            if ( s_selected_timesig.containsKey( s_last_selected_timesig ) ) {
                return s_selected_timesig.get( s_last_selected_timesig );
            } else {
                return null;
            }
        }

        public static int getLastSelectedTimesigBarcount() {
            return s_last_selected_timesig;
        }

        public static void addSelectedTimesig( int barcount ) {
            clearSelectedEvent(); //ここ注意！
            clearSelectedTempo();
            s_last_selected_timesig = barcount;
            if ( !s_selected_timesig.containsKey( barcount ) ) {
                for ( Iterator itr = s_vsq.TimesigTable.iterator(); itr.hasNext(); ){
                    TimeSigTableEntry tte = (TimeSigTableEntry)itr.next();
                    if ( tte.BarCount == barcount ) {
                        s_selected_timesig.put( barcount, new SelectedTimesigEntry( tte, (TimeSigTableEntry)tte.Clone() ) );
                        break;
                    }
                }
            }
        }

        public static void clearSelectedTimesig() {
            s_selected_timesig.clear();
            s_last_selected_timesig = -1;
        }
        #endregion

        #region SelectedTempo
        /// <summary>
        /// 選択されているテンポ変更イベントのリスト
        /// </summary>
        public static TreeMap<Integer, SelectedTempoEntry> getSelectedTempo() {
            return s_selected_tempo;
        }

        public static SelectedTempoEntry getLastSelectedTempo() {
            if ( s_selected_tempo.containsKey( s_last_selected_tempo ) ) {
                return s_selected_tempo.get( s_last_selected_tempo );
            } else {
                return null;
            }
        }

        public static int getLastSelectedTempoClock() {
            return s_last_selected_tempo;
        }

        public static void addSelectedTempo( int clock ) {
            clearSelectedEvent(); //ここ注意！
            clearSelectedTimesig();
            s_last_selected_tempo = clock;
            if ( !s_selected_tempo.containsKey( clock ) ) {
                for ( Iterator itr = s_vsq.TempoTable.iterator(); itr.hasNext(); ){
                    TempoTableEntry tte = (TempoTableEntry)itr.next();
                    if ( tte.Clock == clock ) {
                        s_selected_tempo.put( clock, new SelectedTempoEntry( tte, (TempoTableEntry)tte.Clone() ) );
                        break;
                    }
                }
            }
        }

        public static void clearSelectedTempo() {
            s_selected_tempo.clear();
            s_last_selected_tempo = -1;
        }
        #endregion

        #region SelectedEvent
        public static SelectedEventList getSelectedEvent() {
            return s_selected_event;
        }

        [Obsolete]
        public static SelectedEventList SelectedEvent {
            get {
                return getSelectedEvent();
            }
        }

        public static void removeSelectedEvent( int id ) {
            removeSelectedEventCor( id, false );
        }

        public static void removeSelectedEventSilent( int id ) {
            removeSelectedEventCor( id, true );
        }

        private static void removeSelectedEventCor( int id, boolean silent ) {
            s_selected_event.Remove( id );
            propertyPanel.UpdateValue( s_selected );
        }

        public static void removeSelectedEventRange( int[] ids ) {
            s_selected_event.RemoveRange( ids );
            propertyPanel.UpdateValue( s_selected );
        }

        public static void addSelectedEventRange( int[] ids ) {
            clearSelectedTempo();
            clearSelectedTimesig();
            Vector<Int32> list = new Vector<Int32>( ids );
            VsqEvent[] index = new VsqEvent[ids.Length];
            int count = 0;
            for ( Iterator itr = s_vsq.Track.get( s_selected ).getEventIterator(); itr.hasNext(); ) {
                VsqEvent ev = (VsqEvent)itr.next();
                int find = -1;
                int c = list.size();
                for ( int i = 0; i < c; i++ ) {
                    if ( list.get( i ) == ev.InternalID ) {
                        find = i;
                        break;
                    }
                }
                if ( 0 <= find ) {
                    index[find] = ev;
                    count++;
                }
                if ( count == ids.Length ) {
                    break;
                }
            }
            for ( int i = 0; i < index.Length; i++ ) {
                s_selected_event.Add( new SelectedEventEntry( s_selected, index[i], (VsqEvent)index[i].Clone() ) );
            }
            if ( SelectedEventChanged != null ) {
                SelectedEventChanged( false );
            }
            propertyPanel.UpdateValue( s_selected );
        }

        public static void addSelectedEvent( int id ) {
            addSelectedEventCor( id, false );
        }

        public static void addSelectedEventSilent( int id ) {
            addSelectedEventCor( id, true );
        }

        private static void addSelectedEventCor( int id, boolean silent ) {
            clearSelectedTempo();
            clearSelectedTimesig();
            for ( Iterator itr = s_vsq.Track.get( s_selected ).getEventIterator(); itr.hasNext(); ) {
                VsqEvent ev = (VsqEvent)itr.next();
                if ( ev.InternalID == id ) {
                    s_selected_event.Add( new SelectedEventEntry( s_selected, ev, (VsqEvent)ev.Clone() ) );
                    if ( !silent && SelectedEventChanged != null ) {
                        SelectedEventChanged( false );
                    }
                    break;
                }
            }
            if ( !silent ) {
                propertyPanel.UpdateValue( s_selected );
            }
        }

        public static void clearSelectedEvent() {
            s_selected_event.Clear();
            propertyPanel.UpdateValue( s_selected );
            if ( SelectedEventChanged != null ) {
                SelectedEventChanged( true );
            }
        }
        #endregion

        public static boolean isOverlay() {
            return s_overlay;
        }

        public static void setOverlay( boolean value ) {
            s_overlay = value;
        }

        public static boolean getRenderRequired( int track ) {
            return s_render_required[track - 1];
        }

        public static void setRenderRequired( int track, boolean value ) {
            s_render_required[track - 1] = value;
        }

        /// <summary>
        /// 現在の編集モード
        /// </summary>
        public static EditMode getEditMode() {
            return s_edit_mode;
        }

        public static void setEditMode( EditMode value ) {
            s_edit_mode = value;
        }

        /// <summary>
        /// グリッドを表示するか否かを表すフラグを取得または設定します
        /// </summary>
        public static boolean isGridVisible() {
            return s_grid_visible;
        }

        public static void setGridVisible( boolean value ) {
            if ( value != s_grid_visible ) {
                s_grid_visible = value;
                if ( GridVisibleChanged != null ) {
                    GridVisibleChanged( typeof( AppManager ), new EventArgs() );
                }
            }
        }

        /// <summary>
        /// 現在のプレビューがリピートモードであるかどうかを表す値を取得または設定します
        /// </summary>
        public static boolean isRepeatMode() {
            return s_repeat_mode;
        }

        public static void setRepeatMode( boolean value ) {
            s_repeat_mode = value;
        }

        /// <summary>
        /// 現在プレビュー中かどうかを示す値を取得または設定します
        /// </summary>
        public static boolean isPlaying() {
            return s_playing;
        }

        public static void setPlaying( boolean value ) {
            boolean previous = s_playing;
            s_playing = value;
            if ( previous != s_playing ) {
                if ( s_playing && PreviewStarted != null ) {
                    int clock = getCurrentClock();
                    PreviewStarted( typeof( AppManager ), new EventArgs() );
                } else if ( !s_playing && PreviewAborted != null ) {
                    PreviewAborted( typeof( AppManager ), new EventArgs() );
                }
            }
        }

        /// <summary>
        /// _vsq_fileにセットされたvsqファイルの名前を取得します。
        /// </summary>
        public static String getFileName() {
            return s_file;
        }

        public static void saveTo( String file ) {
            if ( s_vsq != null ) {
                String path = Path.GetDirectoryName( file );
                String file2 = Path.Combine( path, Path.GetFileNameWithoutExtension( file ) + ".vsq" );
                s_vsq.writeAsXml( file );
                s_vsq.write( file2 );
                s_file = file;
                editorConfig.PushRecentFiles( s_file );
            }
        }

        /// <summary>
        /// 位置clockにおけるテンポ、拍子、小節数を取得します
        /// </summary>
        /// <param name="clock"></param>
        /// <param name="numerator"></param>
        /// <param name="denominator"></param>
        /// <param name="bar_count"></param>
        /// <param name="tempo"></param>
        public static void getTempoAndTimeSignature(
            int clock,
            out int numerator,
            out int denominator,
            out int bar_count,
            out int tempo,
            out int time_sig_index,
            out int tempo_index,
            out int last_clock ) {
            numerator = 4;
            denominator = 4;
            bar_count = 0;
            tempo = 480000;
            time_sig_index = -1;
            tempo_index = -1;
            last_clock = 0;
            if ( s_vsq != null ) {
                int index = -1;
                for ( int i = 0; i < s_vsq.TimesigTable.size(); i++ ) {
                    if ( s_vsq.TimesigTable.get( i ).Clock > clock ) {
                        index = i - 1;
                        break;
                    }
                }
                if ( index >= 0 ) {
                    denominator = s_vsq.TimesigTable.get( index ).Denominator;
                    numerator = s_vsq.TimesigTable.get( index ).Numerator;
                    time_sig_index = index;
                    last_clock = s_vsq.TimesigTable.get( index ).Clock;
                }
                bar_count = s_vsq.getBarCountFromClock( clock );

                index = -1;
                for ( int i = 0; i < s_vsq.TempoTable.size(); i++ ) {
                    if ( s_vsq.TempoTable.get( i ).Clock > clock ) {
                        index = i - 1;
                        break;
                    }
                }
                if ( index >= 0 ) {
                    tempo = s_vsq.TempoTable.get( index ).Tempo;
                    tempo_index = index;
                }
            }
        }

        /// <summary>
        /// 現在の演奏マーカーの位置を取得または設定します。
        /// </summary>
        public static int getCurrentClock() {
            return s_current_clock;
        }

        public static void setCurrentClock( int value ) {
            int old = s_current_clock;
            s_current_clock = value;
            int barcount = s_vsq.getBarCountFromClock( s_current_clock );
            int bar_top_clock = s_vsq.getClockFromBarCount( barcount );
            int numerator = 4;
            int denominator = 4;
            s_vsq.getTimesigAt( s_current_clock, out numerator, out denominator );
            int clock_per_beat = 480 / 4 * denominator;
            int beat = (s_current_clock - bar_top_clock) / clock_per_beat;
            s_current_play_position.BarCount = barcount - s_vsq.getPreMeasure() + 1;
            s_current_play_position.Beat = beat + 1;
            s_current_play_position.Clock = s_current_clock - bar_top_clock - clock_per_beat * beat;
            s_current_play_position.Denominator = denominator;
            s_current_play_position.Numerator = numerator;
            s_current_play_position.Tempo = s_vsq.getTempoAt( s_current_clock );
            if ( old != s_current_clock && CurrentClockChanged != null ) {
                CurrentClockChanged( typeof( AppManager ), new EventArgs() );
            }
        }

        /// <summary>
        /// 現在の演奏カーソルの位置(m_current_clockと意味は同じ。CurrentClockが変更されると、自動で更新される)
        /// </summary>
        public static PlayPositionSpecifier getPlayPosition() {
            return s_current_play_position;
        }

        /// <summary>
        /// 現在選択されているトラックを取得または設定します
        /// </summary>
        public static int getSelected() {
            int tracks = s_vsq.Track.size();
            if ( tracks <= s_selected ) {
                s_selected = tracks - 1;
            }
            return s_selected;
        }

        public static void setSelected( int value ) {
            s_selected = value;
        }

        [Obsolete]
        public static int Selected {
            get {
                return getSelected();
            }
        }

        /// <summary>
        /// vsqファイルを読込みます
        /// </summary>
        /// <param name="file"></param>
        public static void readVsq( String file ) {
            s_selected = 1;
            s_file = file;
            VsqFileEx newvsq = null;
            try {
                newvsq = VsqFileEx.readFromXml( file );
            } catch ( Exception ex ) {
#if DEBUG
                AppManager.debugWriteLine( "EditorManager.ReadVsq; ex=" + ex );
#endif
                return;
            }
            if ( newvsq == null ) {
                return;
            }
            s_vsq = newvsq;
            for ( int i = 0; i < s_render_required.Length; i++ ) {
                if ( i < s_vsq.Track.size() - 1 ) {
                    s_render_required[i] = true;
                } else {
                    s_render_required[i] = false;
                }
            }
            startMarker = s_vsq.getPreMeasureClocks();
            int bar = s_vsq.getPreMeasure() + 1;
            endMarker = s_vsq.getClockFromBarCount( bar );
            if ( s_vsq.Track.size() >= 1 ) {
                s_selected = 1;
            } else {
                s_selected = -1;
            }
        }

#if !TREECOM
        /// <summary>
        /// vsqファイル。
        /// </summary>
        public static VsqFileEx getVsqFile() {
            return s_vsq;
        }

        [Obsolete]
        public static VsqFileEx VsqFile {
            get {
                return getVsqFile();
            }
        }
#endif

        public static void setVsqFile( VsqFileEx vsq ) {
            s_vsq = vsq;
            for ( int i = 0; i < s_render_required.Length; i++ ) {
                if ( i < s_vsq.Track.size() - 1 ) {
                    s_render_required[i] = true;
                } else {
                    s_render_required[i] = false;
                }
            }
            s_file = "";
            startMarker = s_vsq.getPreMeasureClocks();
            int bar = s_vsq.getPreMeasure() + 1;
            endMarker = s_vsq.getClockFromBarCount( bar );
        }
        #endregion

        public static void init() {
            loadConfig();
            VSTiProxy.Init();
            s_locker = new object();
            SymbolTable.loadDictionary();
            VSTiProxy.CurrentUser = "";

            #region Apply User Dictionary Configuration
            Vector<ValuePair<String, boolean>> current = new Vector<ValuePair<String, boolean>>();
            for ( int i = 0; i < SymbolTable.getCount(); i++ ) {
                current.add( new ValuePair<String, boolean>( SymbolTable.getSymbolTable( i ).getName(), false ) );
            }
            Vector<ValuePair<String, boolean>> config_data = new Vector<ValuePair<String, boolean>>();
            for ( int i = 0; i < editorConfig.UserDictionaries.size(); i++ ) {
                String[] spl = editorConfig.UserDictionaries.get( i ).Split( "\t".ToCharArray(), 2 );
                config_data.add( new ValuePair<String, boolean>( spl[0], (spl[1].Equals( "T" ) ? true : false) ) );
#if DEBUG
                AppManager.debugWriteLine( "    " + spl[0] + "," + spl[1] );
#endif
            }
            Vector<KeyValuePair<String, boolean>> common = new Vector<KeyValuePair<String, boolean>>();
            for ( int i = 0; i < config_data.size(); i++ ) {
                for ( int j = 0; j < current.size(); j++ ) {
                    if ( config_data.get( i ).Key.Equals( current.get( j ).Key ) ) {
                        current.get( j ).Value = true; //こっちのbooleanは、AppManager.EditorConfigのUserDictionariesにもKeyが含まれているかどうかを表すので注意
                        common.add( new KeyValuePair<String, boolean>( config_data.get( i ).Key, config_data.get( i ).Value ) );
                        break;
                    }
                }
            }
            for ( int i = 0; i < current.size(); i++ ) {
                if ( !current.get( i ).Value ) {
                    common.add( new KeyValuePair<String, boolean>( current.get( i ).Key, false ) );
                }
            }
            SymbolTable.changeOrder( common.toArray( new KeyValuePair<String, boolean>[] { } ) );
            #endregion

            Boare.Lib.AppUtil.Messaging.LoadMessages();
            Boare.Lib.AppUtil.Messaging.Language = editorConfig.Language;

            KeySoundPlayer.Init();
            PaletteToolServer.Init();

#if !TREECOM
            s_id = bocoree.misc.getmd5( DateTime.Now.ToBinary().ToString() );
            String log = Path.Combine( getTempWaveDir(), "run.log" );
#endif
            propertyPanel = new PropertyPanel();
            propertyWindow = new FormNoteProperty();
            propertyWindow.Controls.Add( propertyPanel );
            propertyPanel.Dock = DockStyle.Fill;
        }

        public static String getShortcutDisplayString( Keys[] keys ) {
            String ret = "";
            Vector<Keys> list = new Vector<Keys>( keys );
            if ( list.contains( Keys.Menu ) ) {
                ret = new String( '\x2318', 1 );
            }
            if ( list.contains( Keys.Control ) ) {
                ret += (ret.Equals( "" ) ? "" : "+") + "Ctrl";
            }
            if ( list.contains( Keys.Shift ) ) {
                ret += (ret.Equals( "" ) ? "" : "+") + "Shift";
            }
            if ( list.contains( Keys.Alt ) ) {
                ret += (ret.Equals( "" ) ? "" : "+") + "Alt";
            }
            Vector<Keys> list2 = new Vector<Keys>();
            foreach ( Keys key in keys ) {
                AppManager.debugWriteLine( "    " + key );
                if ( key != Keys.Control && key != Keys.Shift && key != Keys.Alt ) {
                    list2.add( key );
                }
            }
            Collections.sort( list2 );
            for ( int i = 0; i < list2.size(); i++ ) {
                ret += (ret.Equals( "" ) ? "" : "+") + getKeyDisplayString( list2.get( i ) );
            }
            return ret;
        }

        private static String getKeyDisplayString( Keys key ) {
            switch ( key ) {
                case Keys.PageDown:
                    return "PgDn";
                case Keys.PageUp:
                    return "PgUp";
                case Keys.D0:
                    return "0";
                case Keys.D1:
                    return "1";
                case Keys.D2:
                    return "2";
                case Keys.D3:
                    return "3";
                case Keys.D4:
                    return "4";
                case Keys.D5:
                    return "5";
                case Keys.D6:
                    return "6";
                case Keys.D7:
                    return "7";
                case Keys.D8:
                    return "8";
                case Keys.D9:
                    return "9";
                case Keys.Menu:
                    return new String( '\x2318', 1 );
                default:
                    return key.ToString();
            }
        }

        /// <summary>
        /// オブジェクトをシリアライズし，クリップボードに格納するための文字列を作成します
        /// </summary>
        /// <param name="obj"></param>
        /// <returns></returns>
        private static String getSerializedText( object obj ) {
            String str = "";
            using ( MemoryStream ms = new MemoryStream() ) {
                BinaryFormatter bf = new BinaryFormatter();
                bf.Serialize( ms, obj );
                ms.Seek( 0, SeekOrigin.Begin );
                byte[] arr = new byte[ms.Length];
                ms.Read( arr, 0, arr.Length );
                str = CLIP_PREFIX + ":" + obj.GetType().FullName + ":" + Convert.ToBase64String( arr );
            }
            return str;
        }

        /// <summary>
        /// クリップボードに格納された文字列を元に，デシリアライズされたオブジェクトを取得します
        /// </summary>
        /// <param name="s"></param>
        /// <returns></returns>
        private static object getDeserializedObjectFromText( String s ) {
            if ( s.StartsWith( CLIP_PREFIX ) ) {
                int index = s.IndexOf( ":" );
                index = s.IndexOf( ":", index + 1 );
                object ret = null;
                try {
                    using ( MemoryStream ms = new MemoryStream( Convert.FromBase64String( s.Substring( index + 1 ) ) ) ) {
                        BinaryFormatter bf = new BinaryFormatter();
                        ret = bf.Deserialize( ms );
                    }
                } catch {
                    ret = null;
                }
                return ret;
            } else {
                return null;
            }
        }

        public static void clearClipBoard() {
            if ( Clipboard.ContainsText() ) {
                if ( Clipboard.GetText().StartsWith( CLIP_PREFIX ) ) {
                    Clipboard.Clear();
                }
            }
        }

        public static void setClipboard( ClipboardEntry item ) {
            String clip = getSerializedText( item );
            Clipboard.Clear();
            Clipboard.SetText( clip );
        }

        #region VsqEvent用のクリップボード管理
        public static Vector<VsqEvent> getCopiedEvent() {
            if ( Clipboard.ContainsText() ) {
                String clip = Clipboard.GetText();
                if ( clip.StartsWith( CLIP_PREFIX ) ) {
                    int index1 = clip.IndexOf( ":" );
                    int index2 = clip.IndexOf( ":", index1 + 1 );
                    String typename = clip.Substring( index1 + 1, index2 - index1 - 1 );
#if DEBUG
                    AppManager.debugWriteLine( "typename=" + typename );
#endif
                    if ( typename.Equals( typeof( ClipboardEntry ).FullName ) ) {
                        Vector<VsqEvent> ret = null;
                        try {
                            ClipboardEntry ce = (ClipboardEntry)getDeserializedObjectFromText( clip );
                            ret = ce.events;
                        } catch {
                            ret = null;
                        }
                        if ( ret != null ) {
                            return ret;
                        }
                    }
                }
            }
            return new Vector<VsqEvent>();
        }

        public static void setCopiedEvent( Vector<VsqEvent> item ) {
            ClipboardEntry ce = new ClipboardEntry();
            ce.events = item;
            String clip = getSerializedText( ce );
            Clipboard.Clear();
            Clipboard.SetText( clip );
        }
        #endregion

        #region TempoTableEntry用のクリップボード管理
        public static Vector<TempoTableEntry> getCopiedTempo( out int copy_started_clock ) {
            Vector<TempoTableEntry> tempo_table = null;
            copy_started_clock = 0;
            if ( Clipboard.ContainsText() ) {
                String clip = Clipboard.GetText();
                if ( clip.StartsWith( CLIP_PREFIX ) ) {
                    int index1 = clip.IndexOf( ":" );
                    int index2 = clip.IndexOf( ":", index1 + 1 );
                    String typename = clip.Substring( index1 + 1, index2 - index1 - 1 );
                    if ( typename.Equals( typeof( ClipboardEntry ).FullName ) ) {
                        try {
                            ClipboardEntry ce = (ClipboardEntry)getDeserializedObjectFromText( clip );
                            tempo_table = ce.tempo;
                            copy_started_clock = ce.copyStartedClock;
                        } catch {
                            tempo_table = null;
                        }
                    }
                }
            }
            if ( tempo_table == null ) {
                tempo_table = new Vector<TempoTableEntry>();
                copy_started_clock = 0;
            }
            return tempo_table;
        }

        public static void setCopiedTempo( Vector<TempoTableEntry> item, int copy_started_clock ) {
            ClipboardEntry ce = new ClipboardEntry();
            ce.tempo = item;
            String clip = getSerializedText( ce );
            Clipboard.Clear();
            Clipboard.SetText( clip );
        }
        #endregion

        #region TimeSigTableEntry用のクリップボード管理
        public static Vector<TimeSigTableEntry> getCopiedTimesig( out int copy_started_clock ) {
            Vector<TimeSigTableEntry> tempo_table = null;
            copy_started_clock = 0;
            if ( Clipboard.ContainsText() ) {
                String clip = Clipboard.GetText();
                if ( clip.StartsWith( CLIP_PREFIX ) ) {
                    int index1 = clip.IndexOf( ":" );
                    int index2 = clip.IndexOf( ":", index1 + 1 );
                    String typename = clip.Substring( index1 + 1, index2 - index1 - 1 );
                    if ( typename.Equals( typeof( ClipboardEntry ).FullName ) ) {
                        try {
                            ClipboardEntry ce = (ClipboardEntry)getDeserializedObjectFromText( clip );
                            tempo_table = ce.timesig;
                            copy_started_clock = ce.copyStartedClock;
                        } catch {
                            tempo_table = null;
                        }
                    }
                }
            }
            if ( tempo_table == null ) {
                tempo_table = new Vector<TimeSigTableEntry>();
                copy_started_clock = 0;
            }
            return tempo_table;
        }

        public static void setCopiedTimesig( Vector<TimeSigTableEntry> item, int copy_started_clock ) {
            ClipboardEntry ce = new ClipboardEntry();
            ce.timesig = item;
            String clip = getSerializedText( ce );
            Clipboard.Clear();
            Clipboard.SetText( clip );
        }
        #endregion

        #region CurveType, Vector<BPPair>用のクリップボード管理
        public static TreeMap<CurveType, Vector<BPPair>> getCopiedCurve( out int copy_started_clock ) {
            TreeMap<CurveType, Vector<BPPair>> tempo_table = null;
            copy_started_clock = 0;
            if ( Clipboard.ContainsText() ) {
                String clip = Clipboard.GetText();
                if ( clip.StartsWith( CLIP_PREFIX ) ) {
                    int index1 = clip.IndexOf( ":" );
                    int index2 = clip.IndexOf( ":", index1 + 1 );
                    String typename = clip.Substring( index1 + 1, index2 - index1 - 1 );
                    if ( typename.Equals( typeof( ClipboardEntry ).FullName ) ) {
                        try {
                            ClipboardEntry ce = (ClipboardEntry)getDeserializedObjectFromText( clip );
                            tempo_table = ce.curves;
                            copy_started_clock = ce.copyStartedClock;
                        } catch {
                            tempo_table = null;
                        }
                    }
                }
            }
            if ( tempo_table == null ) {
                tempo_table = new TreeMap<CurveType, Vector<BPPair>>();
                copy_started_clock = 0;
            }
            return tempo_table;
        }

        public static void setCopiedCurve( TreeMap<CurveType, Vector<BPPair>> item, int copy_started_clock ) {
            ClipboardEntry ce = new ClipboardEntry();
            ce.curves = item;
            String clip = getSerializedText( ce );
            Clipboard.Clear();
            Clipboard.SetText( clip );
        }
        #endregion

        #region CurveType, Vector<BezierChain>用のクリップボード管理
        public static TreeMap<CurveType, Vector<BezierChain>> getCopiedBezier( out int copy_started_clock ) {
            TreeMap<CurveType, Vector<BezierChain>> tempo_table = null;
            copy_started_clock = 0;
            if ( Clipboard.ContainsText() ) {
                String clip = Clipboard.GetText();
                if ( clip.StartsWith( CLIP_PREFIX ) ) {
                    int index1 = clip.IndexOf( ":" );
                    int index2 = clip.IndexOf( ":", index1 + 1 );
                    String typename = clip.Substring( index1 + 1, index2 - index1 - 1 );
                    if ( typename.Equals( typeof( ClipboardEntry ).FullName ) ) {
                        try {
                            ClipboardEntry ce = (ClipboardEntry)getDeserializedObjectFromText( clip );
                            tempo_table = ce.beziers;
                            copy_started_clock = ce.copyStartedClock;
                        } catch {
                            tempo_table = null;
                        }
                    }
                }
            }
            if ( tempo_table == null ) {
                tempo_table = new TreeMap<CurveType, Vector<BezierChain>>();
                copy_started_clock = 0;
            }
            return tempo_table;
        }

        public static void setCopiedBezier( TreeMap<CurveType, Vector<BezierChain>> item, int copy_started_clock ) {
            ClipboardEntry ce = new ClipboardEntry();
            ce.beziers = item;
            String clip = getSerializedText( ce );
            Clipboard.Clear();
            Clipboard.SetText( clip );
        }
        #endregion

        public static Assembly compileScript( String code, out CompilerResults results ) {
            CSharpCodeProvider provider = new CSharpCodeProvider();
            String path = Application.StartupPath;
            CompilerParameters parameters = new CompilerParameters( new String[] {
                Path.Combine( path, "Boare.Lib.Vsq.dll" ),
                Path.Combine( path, "Cadencii.exe" ),
                Path.Combine( path, "Boare.Lib.Media.dll" ),
                Path.Combine( path, "Boare.Lib.AppUtil.dll" ),
                Path.Combine( path, "bocoree.dll" ) } );
            parameters.ReferencedAssemblies.Add( "System.Windows.Forms.dll" );
            parameters.ReferencedAssemblies.Add( "System.dll" );
            parameters.ReferencedAssemblies.Add( "System.Drawing.dll" );
            parameters.ReferencedAssemblies.Add( "System.Xml.dll" );
            parameters.GenerateInMemory = true;
            parameters.GenerateExecutable = false;
            parameters.IncludeDebugInformation = true;
            try {
                results = provider.CompileAssemblyFromSource( parameters, code );
                return results.CompiledAssembly;
            } catch ( Exception ex ){
#if DEBUG
                AppManager.debugWriteLine( "AppManager.CompileScript; ex=" + ex );
#endif
                results = null;
                return null;
            }
        }

        /// <summary>
        /// アプリケーションデータの保存位置を取得します
        /// Gets the path for application data
        /// </summary>
        public static String getApplicationDataPath() {
            String dir = Path.Combine( Environment.GetFolderPath( Environment.SpecialFolder.LocalApplicationData ), "Boare" );
            if ( !Directory.Exists( dir ) ) {
                Directory.CreateDirectory( dir );
            }
            String dir2 = Path.Combine( dir, CONFIG_DIR_NAME );
            if ( !Directory.Exists( dir2 ) ) {
                Directory.CreateDirectory( dir2 );
            }
            return dir2;
        }

        /// <summary>
        /// 位置クオンタイズ時の音符の最小単位を、クロック数に換算したものを取得します
        /// </summary>
        /// <returns></returns>
        public static int getPositionQuantizeClock() {
            return QuantizeModeUtil.GetQuantizeClock( editorConfig.PositionQuantize, editorConfig.PositionQuantizeTriplet );
        }

        /// <summary>
        /// 音符長さクオンタイズ時の音符の最小単位を、クロック数に換算したものを取得します
        /// </summary>
        /// <returns></returns>
        public static int getLengthQuantizeClock() {
            return QuantizeModeUtil.GetQuantizeClock( editorConfig.LengthQuantize, editorConfig.LengthQuantizeTriplet );
        }

        public static void saveConfig() {
            // ユーザー辞書の情報を取り込む
            editorConfig.UserDictionaries.clear();
            for ( int i = 0; i < SymbolTable.getCount(); i++ ) {
                editorConfig.UserDictionaries.add( SymbolTable.getSymbolTable( i ).getName() + "\t" + (SymbolTable.getSymbolTable( i ).isEnabled() ? "T" : "F") );
            }

            String file = Path.Combine( getApplicationDataPath(), CONFIG_FILE_NAME );
            try {
                EditorConfig.Serialize( editorConfig, file );
            } catch {
            }
        }

        public static void loadConfig() {
            String config_file = Path.Combine( getApplicationDataPath(), CONFIG_FILE_NAME );
            EditorConfig ret = null;
            if ( File.Exists( config_file ) ) {
                try {
                    ret = EditorConfig.Deserialize( editorConfig, config_file );
                } catch {
                    ret = null;
                }
            } else {
                config_file = Path.Combine( Application.StartupPath, CONFIG_FILE_NAME );
                if ( File.Exists( config_file ) ) {
                    try {
                        ret = EditorConfig.Deserialize( editorConfig, config_file );
                    } catch {
                        ret = null;
                    }
                }
            }
            if ( ret == null ) {
                ret = new EditorConfig();
            }
            editorConfig = ret;
            for ( int i = 0; i < SymbolTable.getCount(); i++ ) {
                SymbolTable st = SymbolTable.getSymbolTable( i );
                boolean found = false;
                for( Iterator itr = editorConfig.UserDictionaries.iterator(); itr.hasNext() ; ){
                    String s = (String)itr.next();
                    String[] spl = s.Split( "\t".ToCharArray(), 2 );
                    if ( st.getName().Equals( spl[0] ) ) {
                        found = true;
                        break;
                    }
                }
                if ( !found ) {
                    editorConfig.UserDictionaries.add( st.getName() + "\tT" );
                }
            }
            MidiPlayer.DeviceGeneral = (uint)editorConfig.MidiDeviceGeneral.PortNumber;
            MidiPlayer.DeviceMetronome = (uint)editorConfig.MidiDeviceMetronome.PortNumber;
            MidiPlayer.NoteBell = editorConfig.MidiNoteBell;
            MidiPlayer.NoteNormal = editorConfig.MidiNoteNormal;
            MidiPlayer.PreUtterance = editorConfig.MidiPreUtterance;
            MidiPlayer.ProgramBell = editorConfig.MidiProgramBell;
            MidiPlayer.ProgramNormal = editorConfig.MidiProgramNormal;
            MidiPlayer.RingBell = editorConfig.MidiRingBell;
        }

        public static VsqID getSingerIDUtau( String name ) {
            VsqID ret = new VsqID( 0 );
            ret.type = VsqIDType.Singer;
            int index = -1;
            for ( int i = 0; i < editorConfig.UtauSingers.size(); i++ ) {
                if ( editorConfig.UtauSingers.get( i ).VOICENAME.Equals( name ) ) {
                    index = i;
                    break;
                }
            }
            if ( index >= 0 ) {
                SingerConfig sc = editorConfig.UtauSingers.get( index );
                int lang = 0;//utauは今のところ全部日本語
                ret.IconHandle = new IconHandle();
                ret.IconHandle.IconID = "$0701" + index.ToString( "0000" );
                ret.IconHandle.IDS = sc.VOICENAME;
                ret.IconHandle.Index = 0;
                ret.IconHandle.Language = lang;
                ret.IconHandle.Length = 1;
                ret.IconHandle.Original = sc.Original;
                ret.IconHandle.Program = sc.Program;
                ret.IconHandle.Caption = "";
                return ret;
            } else {
                ret.IconHandle = new IconHandle();
                ret.IconHandle.Program = 0;
                ret.IconHandle.Language = 0;
                ret.IconHandle.IconID = "$0701" + 0.ToString( "0000" );
                ret.IconHandle.IDS = "Unknown";
                ret.type = VsqIDType.Singer;
                return ret;
            }
        }

        public static SingerConfig getSingerInfoUtau( int program_change ) {
            if ( 0 <= program_change && program_change < editorConfig.UtauSingers.size() ) {
                return editorConfig.UtauSingers.get( program_change );
            } else {
                return null;
            }
        }

        public static String getVersion() {
            String prefix = "";
            String rev = "";
            // $Id: AppManager.cs 420 2009-08-17 16:28:38Z kbinani $
            String id = getAssemblyConfigurationAttribute();
            String[] spl0 = id.Split( " ".ToCharArray(), StringSplitOptions.RemoveEmptyEntries );
            if ( spl0.Length >= 3 ) {
                String s = spl0[2];
#if DEBUG
                AppManager.debugWriteLine( "AppManager.get__VERSION; s=" + s );
#endif
                String[] spl = s.Split( " ".ToCharArray(), StringSplitOptions.RemoveEmptyEntries );
                if ( spl.Length > 0 ) {
                    rev = spl[0];
                }
            }
            if ( rev.Equals( "" ) ) {
                rev = "?";
            }
#if DEBUG
            prefix = "(rev: " + rev + "; build: debug)";
#else
                prefix = "(rev: " + rev + "; build: release)";
#endif
            return getAssemblyFileVersion( typeof( AppManager ) ) + " " + prefix;
        }

        public static String getAssemblyConfigurationAttribute() {
            Assembly a = Assembly.GetAssembly( typeof( AppManager ) );
            AssemblyConfigurationAttribute attr = (AssemblyConfigurationAttribute)Attribute.GetCustomAttribute( a, typeof( AssemblyConfigurationAttribute ) );
#if DEBUG
            AppManager.debugWriteLine( "GetAssemblyConfigurationAttribute; attr.Configuration=" + attr.Configuration );
#endif
            return attr.Configuration;
        }

        public static String getAssemblyFileVersion( Type t ) {
            Assembly a = Assembly.GetAssembly( t );
            AssemblyFileVersionAttribute afva = (AssemblyFileVersionAttribute)Attribute.GetCustomAttribute( a, typeof( AssemblyFileVersionAttribute ) );
            return afva.Version;
        }

        public static String getAssemblyNameAndFileVersion( Type t ) {
            Assembly a = Assembly.GetAssembly( t );
            AssemblyFileVersionAttribute afva = (AssemblyFileVersionAttribute)Attribute.GetCustomAttribute( a, typeof( AssemblyFileVersionAttribute ) );
            return a.GetName().Name + " v" + afva.Version;
        }

        public static SolidBrush getHilightBrush() {
            return s_hilight_brush;
        }

        public static Color getHilightColor() {
            return s_hilight_brush.Color;
        }

        public static void setHilightColor( Color value ) {
            s_hilight_brush = new SolidBrush( value );
        }

        /// <summary>
        /// ベースとなるテンポ。
        /// </summary>
        public static int getBaseTempo() {
            return s_base_tempo;
        }

        public static void setBaseTempo( int value ) {
            s_base_tempo = value;
        }
    }

}
